% 'Matchine' software by Joan Alexis Glaunes, Copyright (C) Universite Paris Descartes, All Rights Reserved 2005-2006, version 1.0
%
% Any use of this code should make reference to at least one of the following references:
%
% - [for the use of 'measure matching']:
% J. Glaunes, L. Younes and A. Trouve, Diffeomorphic matching of distributions: A new approach for unlabelled point-sets and sub-manifolds matching, Proc. of the 2004 IEEE Computer Society Conference on Computer Vision and Pattern Recognition. (CVPR'04), Vol. 2, pp. 712--718, DOI: 10.1109/CVPR.2004.81
%
% - [for the use of 'curve matching']:
% J. Glaunes, A. Qiu, M. Miller, L. Younes, Large Deformation Diffeomorphic Metric Curve Mapping, International Journal of Computer vision, Springer, 2008, Vol. 80, No. 3, pp. 317--336, DOI: 10.1007/s11263-008-0141-9
%
% - [for the use of 'surface matching']:
% M. Vaillant and J. Glaunes, Surface Matching via Currents, Proc. of Information Processing in Medical Imaging (IPMI'05), Lecture Notes in Computer Science vol. 3565, Springer 2005, pp. 381--392


function s = curvecurr(s)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%   Curve matching via currents - to be used with match.m %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% contact : http://www.mi.parisdescartes.fr/~glaunes/

method = 'curvecurr';

% y (target points) is required
if ~isfield(s,'y')
    error('input structure must contain field <y> (target points)');
end

% y must be 3D data
if size(s.y,1)~=3
    error('field <y> must be 3-rows matrix');
end

y = s.y;
ny = size(y,2);

% vx and vy (template and target faces) are required components
if isfield(s,'vx')~=1 | isfield(s,'vy')~=1
    error('input structure must contain fields vx and vy (template and target faces)');
end

vx = s.vx;
vy = s.vy;
nfx = size(vx,2);         % number of template faces
nfy = size(vy,2);         % number of target faces

numbminims = 1;
sigmaW = 'auto';            % matching kernel size
decsigmaW = 2.5;            % decrease rate of sigmaW after each minimization

usefgt = 0;                 % if 1 use IFGT algorithm
testfgt = 0;
order = 5;                 % IFGT order
K = min([10,floor(ny/2)]);  % IFGT number of clusters
e = 9;                      % IFGT cutoff ratio for targets
wx = ones(1,nfx);
wy = ones(1,nfy);

% update by input structure
loadstruct('s',who)

% auto scale
sclsz = mean(max(y')-min(y'));
if strcmp(sigmaW,'auto')
    sigmaW = .5*sclsz;
end
sigmaW2 = sigmaW^2;

% computation of centers and normals for each face of target
[cy,tauy] = compcurr(y,vy,wy);

premin = @funpremin;
matching = @funmatching;
gradmatching = @fungradmatching;

clear s
savestruct('s',who);

% multiscale settings
sigmaW = sigmaW*decsigmaW^numbminims;   %%% numbminims should be given by match.m
sigmaW2 = sigmaW^2;

if usefgt
    WdScalProd = @fgtWdScalProd;
    gradWdNorm = @fgtgradWdNorm;
else
    WdScalProd = @regWdScalProd;
    gradWdNorm = @reggradWdNorm;
end

% process this before each minimization
    function funpremin
        sigmaW = sigmaW/decsigmaW;
        sigmaW2 = sigmaW^2;
        if usefgt*testfgt
            essdir = regWdScalProd(cy,tauy,cy,tauy,sigmaW2);
            essfgt = fgtWdScalProd(cy,tauy,cy,tauy,sigmaW2);
            disp(['relative error fgt curvecurr: ',num2str(abs((essdir-essfgt)/essdir))])
        end
    end

    function R = funmatching(phix)
        % Function:  compute energy
        %
        % Input:
        %        phix     coordinates of vertices on the first curve
        % Output:
        %        R        total current energy
        [cphix,tauphix] = compcurr(phix,vx,wx); % computes centers and normals for each face
        R = WdScalProd([cphix,cy],[tauphix,-tauy],[cphix,cy],[tauphix,-tauy],sigmaW2);
    end


    function g = fungradmatching(phix)
        [cphix,tauphix] = compcurr(phix,vx,wx); % computation of centers and normals for each face
        g = gradWdNorm(phix,cphix,tauphix,sigmaW2);
    end


    function [ca,taua] = compcurr(a,ba,wa)
        %Function:
        %      computer segments' centers and tangents
        %
        % Input:
        %     a    coordinates of vertices, matrix  3*number of vertices
        %     ba   segments'connectivity, matrix 3 * number of segments
        % Output:
        %     ca   coordinates of center for each segment, array  3*number of
        %     segments
        %     taua   tangent directions for each segment, array 3*number of segments
        %
        % Variables:
        %      na   number of vertices
        %      nfa  number of segments

        nfa = size(ba,2);
        ca = zeros(3,nfa);
        taua = zeros(3,nfa);

        % computation of centers and normals for each face
        for f=1:nfa,
            b1 = 3*(ba(1,f)-1);
            b2 = 3*(ba(2,f)-1);
            ca(1,f) = (a(1+b1)+a(1+b2))/2;
            ca(2,f) = (a(2+b1)+a(2+b2))/2;
            ca(3,f) = (a(3+b1)+a(3+b2))/2;
            taua(1,f) = wa(f) * (a(1+b2)-a(1+b1));
            taua(2,f) = wa(f) * (a(2+b2)-a(2+b1));
            taua(3,f) = wa(f) * (a(3+b2)-a(3+b1));
        end

    end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Scalar product in W^* space %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Function:
%             compute currents in W* space
% Input
%         ca      center of each triangle on the first surface
%         taua      normal of each triangle on the first surface
%         cb      center of each triangle on the second surface
%         taub      normal of each triangle on the second surface
% Output
%        R        currnets
%

    function R = regWdScalProd(ca,taua,cb,taub,sigmaW2)
        nfa = size(ca,2);
        nfb = size(cb,2);
        R = 0;
        for k = 1:nfa
            for j = 1:nfb
                argin = -((ca(1,k)-cb(1,j))^2+(ca(2,k)-cb(2,j))^2+(ca(3,k)-cb(3,j))^2)/sigmaW2;
argout = exp(argin);  %% BUILT IN KERNEL kerI, do not remove this comment
                R = R + argout * (taua(1,k)*taub(1,j)+taua(2,k)*taub(2,j)+taua(3,k)*taub(3,j));
            end
        end
    end

    function R = fgtWdScalProd(ca,taua,cb,taub,sigmaW2)
        R = sum(sum(taub.*fgt(3,ca,taua,cb,sqrt(sigmaW2),order,K,e)));
    end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% gradient of W^* norm %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Function:  gradient of data attachment (curve current)
%
% Input
%        phi     coordinates of vertices on the first surface
%        cphi    coordiantes of the center for each triangle on the first
%        surface
%        tauphi    normals for each triangle on the first surface
%
% Output:
%        eta  gradient of current with respect to the coordinates
%        of vertices
%
% Variables
%        etaf_w       gradient on normals with respect to the coordinates
%        of vertices
%        etaf_dw      gradient on kernel with respect to the coordinates of
%        vertices
%

    function eta = reggradWdNorm(phi,cphi,tauphi,sigmaW2)
        eta = zeros(size(phi));
        etaloc = zeros(3,2);
        etaf_w= zeros(3,1);
        etaf_dw = zeros(3,1);

        for e = 1:nfx
            loce = 3*(e-1);
            etaf_w(1) = 0;
            etaf_w(2) = 0;
            etaf_w(3) = 0;
            etaf_dw(1) = 0;
            etaf_dw(2) = 0;
            etaf_dw(3) = 0;
            if wx(e)
            for f = 1:nfx
                locf = 3*(f-1);
                argin = -(...
                    (cphi(1+locf)-cphi(1+loce))^2 + ...
                    (cphi(2+locf)-cphi(2+loce))^2 + ...
                    (cphi(3+locf)-cphi(3+loce))^2 )/sigmaW2;
argout = exp(argin);  %% BUILT IN KERNEL kerI, do not remove this comment
                etaf_w(1) = etaf_w(1) + argout * tauphi(1+locf);
                etaf_w(2) = etaf_w(2) + argout * tauphi(2+locf);
                etaf_w(3) = etaf_w(3) + argout * tauphi(3+locf);
argout = exp(argin);  %% BUILT IN KERNEL derI, do not remove this comment
                dottau = tauphi(1+locf)*tauphi(1+loce) + tauphi(2+locf)*tauphi(2+loce) + tauphi(3+locf)*tauphi(3+loce);
                etaf_dw(1) = etaf_dw(1) + argout * dottau * (cphi(1+locf)-cphi(1+loce));
                etaf_dw(2) = etaf_dw(2) + argout * dottau * (cphi(2+locf)-cphi(2+loce));
                etaf_dw(3) = etaf_dw(3) + argout * dottau * (cphi(3+locf)-cphi(3+loce));
            end
            for f = 1:nfy
                locf = 3*(f-1);
                argin = -(...
                    (cy(1+locf)-cphi(1+loce))^2 + ...
                    (cy(2+locf)-cphi(2+loce))^2 + ...
                    (cy(3+locf)-cphi(3+loce))^2 )/sigmaW2;
argout = exp(argin);  %% BUILT IN KERNEL kerI, do not remove this comment
                etaf_w(1) = etaf_w(1) - argout * tauy(1+locf);
                etaf_w(2) = etaf_w(2) - argout * tauy(2+locf);
                etaf_w(3) = etaf_w(3) - argout * tauy(3+locf);
argout = exp(argin);  %% BUILT IN KERNEL derI, do not remove this comment
                dottau = tauy(1+locf)*tauphi(1+loce) + tauy(2+locf)*tauphi(2+loce) + tauy(3+locf)*tauphi(3+loce);
                etaf_dw(1) = etaf_dw(1) - argout * dottau * (cy(1+locf)-cphi(1+loce));
                etaf_dw(2) = etaf_dw(2) - argout * dottau * (cy(2+locf)-cphi(2+loce));
                etaf_dw(3) = etaf_dw(3) - argout * dottau * (cy(3+locf)-cphi(3+loce));
            end
            end
            etaf_w(1) = wx(e) * etaf_w(1);
            etaf_w(2) = wx(e) * etaf_w(2);
            etaf_w(3) = wx(e) * etaf_w(3);
            etaf_dw(1) = 2 * etaf_dw(1) / sigmaW2;
            etaf_dw(2) = 2 * etaf_dw(2) / sigmaW2;
            etaf_dw(3) = 2 * etaf_dw(3) / sigmaW2;
            %   for k = 1:3
            %     for j = 1:3
            %       b(k+3*(j-1)) = phi(k+3*(vx(j+loce)-1));
            %     end
            %   end
            etaloc(1) = -2*etaf_w(1) + etaf_dw(1);
            etaloc(2) = -2*etaf_w(2) + etaf_dw(2);
            etaloc(3) = -2*etaf_w(3) + etaf_dw(3);
            etaloc(4) = 2*etaf_w(1) + etaf_dw(1);
            etaloc(5) = 2*etaf_w(2) + etaf_dw(2);
            etaloc(6) = 2*etaf_w(3) + etaf_dw(3);

            loc1 = 3*(vx(1+2*(e-1))-1);
            loc2 = 3*(vx(2+2*(e-1))-1);
            eta(1+loc1) = eta(1+loc1) + etaloc(1);
            eta(2+loc1) = eta(2+loc1) + etaloc(2);
            eta(3+loc1) = eta(3+loc1) + etaloc(3);
            eta(1+loc2) = eta(1+loc2) + etaloc(4);
            eta(2+loc2) = eta(2+loc2) + etaloc(5);
            eta(3+loc2) = eta(3+loc2) + etaloc(6);
        end

    end

    function eta = fgtgradWdNorm(phi,cphi,tauphi,sigmaW2)
        cz = [cphi,cy];
        tauz = [tauphi,-tauy];

        resfgt = fgt(3,cz,[tauz;specprod(cz,tauz,1)],cphi,sigmaW,order,K,e);

        eta_dw = (2/sigmaW2) * (specprod(resfgt(4:end,:),tauphi,3) - specprod(specprod(cphi,tauphi,1),resfgt(1:3,:),3));
        eta_w = 2*resfgt(1:3,:);

        eta = zeros(size(phi));
        for m = 1:nfx
            locm = 3*m-2:3*m;
            bm1 = 3*vx(1,m);
            locbm1 = bm1-2:bm1;
            eta(locbm1) = eta(locbm1)' + eta_dw(locm) - wx(m) * eta_w(locm);
            bm2 = 3*vx(2,m);
            locbm2 = bm2-2:bm2;
            eta(locbm2) = eta(locbm2)' + eta_dw(locm) + wx(m) * eta_w(locm);
        end

    end

end

function C = specprod(A,B,m)

n = size(A,2);
a = size(A,1);
b = size(B,1);

A = reshape(A,a/m,m,n);
B = reshape(B,m,b/m,n);
C = zeros(a/m,b/m,n);

for p = 1:n
    C(:,:,p) = A(:,:,p)*B(:,:,p);
end

C = reshape(C,a*b/m/m,n);

end


